/*
 * 
 * POP3-Virusscan-Proxy 0.4
 *
 * (C) 2002 by Folke Ashberg <folke@ashberg.de>
 *
 * $Id: pop3vscan.h,v 1.12 2002/06/11 23:09:16 folke Exp $ 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _POP3VSCVAN_H
#define _POP3VSCVAN_H

/* includes {{{1 */
#include <syslog.h>	    /* do_log */
#include <arpa/inet.h>	    /* config */

/* definitions {{{1 */
#define PROGNAME		"POP3VScan"
#define VERSION			"0.4"
#define MAX_PSEUDO_ARGV		50
#define MESSAGE_NOVIRINFO	"<no virusinfo could be examined>"

#define CONFIG_STATE_CMDPRE	1
#define CONFIG_STATE_FILE	2
#define CONFIG_STATE_CMD	3

#define SCANNER_INIT_NO		0
#define SCANNER_INIT_OK		1
#define SCANNER_INIT_NULL	2 /* scanner needs no init */
#define SCANNER_INIT_ERR	-1

#define SCANNER_RET_OK		0
#define SCANNER_RET_ERR		-1
#define SCANNER_RET_VIRUS	2

/* macros {{{1 */
#if defined(__USE_FILE_OFFSET64) || defined(__USE_LARGEFILE64)
#define FSTAT(a, b) fstat64(a,b)
#define LSTAT(a, b) lstat64(a,b)
#define STAT(a, b) stat64(a,b)
#define stat_t stat64
#else
#define FSTAT(a, b) fstat(a,b)
#define LSTAT(a, b) lstat(a,b)
#define STAT(a, b) stat(a,b)
#define stat_t stat
#endif

#define SETIFNULL(a,b) if (!a) a=b  /* wow, that's magic */
#define NONULL(x) ( x==NULL ? "" : x) /* this is nice, found in the mutt code */

#define TRIM(a) \
{ \
    int len=0; \
      while ((a)[len] && isspace((a)[len])) len++; \
	memmove((a),&(a)[len], strlen(&(a)[len])+1); \
	  while ((len=strlen(a))>0 && isspace((a)[len-1])) (a)[len-1]='\0'; \
}

/* prototypes {{{1 */
typedef struct proxycontext {
    struct linebuf	*serverbuf;
    struct linebuf	*clientbuf;
    struct sockaddr_in	client_addr;
    struct sockaddr_in	server_addr;
    size_t		socksize;
    int			client_fd;
    int			server_fd;
    int			ismail;
    unsigned int	msgnum;
    int			headerda;
    struct paramlist	*params;
    int			mailcount;
    unsigned long	bytecount;
    char		mailfile[4096];
    char		maildir[4096]; /* mailfile.content */
    char		* scanthis; /* depending on demime linked to mailfile / maildir */
    char		* virinfo; /* has to be filled from the scanner */
    int			scannerinit; /* see SCANNER_INIT_* */
} proxycontext;

typedef struct scanner_t {
    char *name;
    char *descr;
    int (*init1)(void);
    int (*init2)(struct proxycontext *);
    int (*scan)(struct proxycontext *, char ** virname);
    void (*uninit2)(struct proxycontext *);
    void (*uninit1)(void);
    int dirscan;
} scanner_t;


typedef struct configuration_t {
    int	    maxchilds;
    char  * runasuser;
    char  * virusdir;
    char  * virustemplate;
    char  * virusscanner;
    int	    viruscode;
    char  * virusregexp;
    int	    virusregexpsub;
    int	    demime; 
    char  * pidfile;
    char  * syslogname;
    char  * configfile; 
    int	    debug;
    struct sockaddr_in addr;
    scanner_t *scanner;
    int scannerenabled;
} configuration_t;

/* function protos {{{1 */
extern void do_log(int level, const char *fmt,...);

/* }}} */
#endif _POP3VSCAN_H
/* vim:set fdm=marker: */
